function cvodes_pde
%% Main function for CVODES sensitivity integration
%% and generation of impulse response matrix for heat equation (PDE)

clear all
close all
clc
format compact
format short

%% SUNDIALS STARTUP

w = pwd;
sundials_path = '/Applications/MATLAB_R2008a/toolbox/sundialsTB';
cd(sundials_path);
startup_STB();
cd(w);

%% SET OPTIONS (true = 1, false = 0)

run = struct;
run.plot_fig         = 1; % plot solution
run.plot_fig_title   = 1; % title names in figures
run.plot_runtime_fig = 1; % plot runtime figures
run.print_info       = 1; % print information
run.rt_mode          = 2; % choose runtime measurement mode

mondata = struct;       % monitor data struct
mondata.stats = 0;      % report step size and method order
mondata.cntr  = 0;      % report counter evolution
mondata.mode  = 'both'; % graphical/text/both
mondata.sol   = 0;      % plot solution
mondata.sensi = 0;      % plot sensitivities (FSA)

% different modes for measuring runtime
switch run.rt_mode
    case 1
        rt_Nx = 2:6:200;
        pts = 10;
        avg_pts = 10; % number of iterations to average over
    case 2
        rt_Nx = 50;
        pts = 50;
        avg_pts = 1; % number of iterations to average over
    otherwise
        fprintf('ERROR: rt_mode not valid');
end

%% USER DATA / PARAMETERS

global ud;  % user data structure

% time
t0 = 0;
tf = 0.5;
Ts = tf/pts;
t = t0:Ts:tf;
N = length(t);

for avg_iter = 1:avg_pts;
    for rt_iter = 1:length(rt_Nx);

        % system
        Nx = rt_Nx(rt_iter); %must be even
        Nu = 1;
        Nz = Nx; % full state measurement
        states = 1/Nx:1/Nx:1;
        delta = 1/Nx;
        v = [-2 1 zeros(1,Nx-2)];
        ud.A = toeplitz(v,v)/delta^2;
        ud.B = [1 zeros(1,Nx-1)]'/delta^2;

        fprintf(['Iteration ' num2str(rt_iter) '/' num2str(length(rt_Nx)) ...
            ', Average Iteration ' num2str(avg_iter) '/' num2str(avg_pts) '\n']);

        %% CVODES INIT

            options = CVodeSetOptions('RelTol',1.e-6,...
                'AbsTol',1.e-6,...
                'JacobianFn',@cvodes_pde_J);

            options = CVodeSetOptions(options,'MonitorFn',@CVodeMonitor,...
                'MonitorData',mondata,...
                'MaxStep',Ts);

            x0 = zeros(Nx,1);                         % initial values
            CVodeMalloc(@cvodes_pde_f,t0,x0,options); % allocate memory

            %% FORWARD SENSITIVITY ANALYSIS INIT

            FSAoptions = CVodeSetFSAOptions('SensErrControl', 'on',...
                'SensRhsFn', @cvodes_pde_fS);

            xS0 = [eye(Nx) zeros(Nx,Nu)];                               % sensitivity initial values
            CVodeSensMalloc(Nx + Nu, 'Simultaneous', xS0, FSAoptions);  % allocate memory

            %% PROBLEM SOLUTION

            % prealloc for speed
            x = zeros(Nx,N);
            AS    = zeros(Nx,Nx,N-1);
            BS    = zeros(Nx,Nu,N-1);

            % sinus input
            periods = 2;
            freq = periods/tf;
            omega = 2*pi*freq;
            u = sin(omega*t);

            x(:,1) = x0;
            tic;
            for k = 2:N
                % update control
                ud.u = u(k);

                % reinit system equations
                CVodeReInit(@cvodes_pde_f,t(k)-Ts,x(:,k-1),options);

                % reinit sensitivity equations
                CVodeSensReInit('Simultaneous', xS0, FSAoptions);

                % integration
                [status, tk, xk, xSk] = CVode(t(k),'Normal');

                % store data
                x(:,k) = xk;
                AS(:,:,k) = xSk(:,1:Nx);
                BS(:,:,k) = xSk(:,Nx+1:Nx+Nu);
            end
            t_cvodes(rt_iter,avg_iter) = toc;

            si = CVodeGetStats; % store stats

            CVodeFree;  % free memory

            if run.plot_fig == true

                %% FIGURES

                figure()
                mesh(states,t,x')
                if run.plot_fig_title == true
                    title(['solution']);
                end
                axis([0 1 0 tf -1 1])
                title('Zero state response - sinusoidal input')
                xlabel('x')
                ylabel('t')
                zlabel('Temperature')

                figure();
                plot(t,u','-*');
                if run.plot_fig_title == true
                    title(['control']);
                end
                legend('u');
                xlabel('t [s]');
                grid;

            end

            % define matrices X = [X_1 X_2 ... X_N]
            A = AS(:,:,2:N);
            B = BS(:,:,2:N);
            C = zeros(Nx,Nx,N);
            D = zeros(Nz,Nu,N);
            for k = 1:N
                C(:,:,k) = eye(Nx);
            end


            %% BUILD IMPULSE RESPONSE MATRIX - FORWARD METHOD

            % prealloc for speed
            Xi_fwd = zeros(N*Nz,N*Nu);
            phi = zeros(Nx,Nu*(N-1));

            tic;
            Xi_fwd(1:Nz,1:Nu) = D(:,:,1);
            for j=2:N
                for i=1:j-2
                    phi(1:Nx,(i-1)*Nu+1:i*Nu) = A(:,:,j-1)*phi(1:Nx,(i-1)*Nu+1:i*Nu);

                end
                phi(1:Nx,(j-2)*Nu+1:(j-1)*Nu) = B(:,:,j-1);
                for i = 1:j-1
                    Xi_fwd((j-1)*Nz+1:j*Nz,(i-1)*Nu+1:i*Nu) = C(:,:,j)*phi(1:Nx,(i-1)*Nu+1:i*Nu);
                end
                Xi_fwd((j-1)*Nz+1:j*Nz,(j-1)*Nu+1:j*Nu) = D(:,:,j);
            end
            t_fwd(rt_iter,avg_iter) = toc;


            %% BUILD IMPULSE RESPONSE MATRIX - ADJOINT METHOD

            % prealloc for speed
            Xi_adj = zeros(N*Nz,N*Nu);
            lambda = zeros(Nz,Nx*(N-1));
            con_enf = zeros(1,N);

            con_enf(1) = 1;
            con_enf(2) = 1;
            con_enf(3) = 1;

            tic;
            Xi_adj((N-1)*Nz+1:N*Nz,(N-1)*Nu+1:N*Nu) = D(:,:,N);
            for k = N-1:-1:1
                lambda((k-1)*Nz+1:k*Nz,1:Nx) = C(:,:,k+1);
                for i = k+1:N-1
                    lambda((i-1)*Nz+1:i*Nz,1:Nx) = lambda((i-1)*Nz+1:i*Nz,1:Nx)*A(:,:,k+1);
                end
                Xi_adj((k-1)*Nz+1:k*Nz,(k-1)*Nu+1:k*Nu) = D(:,:,k);
                for i = k:N-1
                    Xi_adj(i*Nz+1:(i+1)*Nz,(k-1)*Nu+1:k*Nu) = lambda((i-1)*Nz+1:i*Nz,1:Nx)*B(:,:,k);
                end
            end
            t_adj(rt_iter,avg_iter) = toc;


            %% BUILD IMPULSE RESPONSE MATRIX - ON3

            % prealloc for speed
            Theta_ON3 = zeros(N*Nx,N*Nu);
            Xi_ON3    = zeros(N*Nz,N*Nu);

            tic;
            for k=1:N-1
                for j=k+1:N
                    Theta_ON3((j-1)*Nx+1:j*Nx,(k-1)*Nu+1:k*Nu) = B(:,:,k);
                end

                for i=1:k
                    for j = k+2:N
                        Theta_ON3((j-1)*Nx+1:j*Nx,(i-1)*Nu+1:i*Nu) = ...
                            A(:,:,k+1)*Theta_ON3((j-1)*Nx+1:j*Nx,(i-1)*Nu+1:i*Nu);
                    end
                end
            end

            for k=1:N
                Ck = C(:,:,k);
                Dk = D(:,:,k);
                for i=1:k-1
                    Xi_ON3((k-1)*Nz+1:k*Nz,(i-1)*Nu+1:i*Nu) = ...
                        Ck*Theta_ON3((k-1)*Nx+1:k*Nx,(i-1)*Nu+1:i*Nu);
                end
                Xi_ON3((k-1)*Nz+1:k*Nz,(k-1)*Nu+1:k*Nu) = Dk;
            end
            t_ON3(rt_iter,avg_iter) = toc;


            if run.print_info == true

                %% PRINT INFORMATION

                fprintf(['INFORMATION\n']);

                fprintf(['Start time: ' num2str(t0) '\n']);
                fprintf(['Sampling time: ' num2str(Ts) '\n']);
                fprintf(['Stop time: ' num2str(tf) '\n']);
                fprintf(['Number of sampling points: ' num2str(N) '\n\n']);

                fprintf(['CVODES integration time: ' num2str(t_cvodes(rt_iter,avg_iter)) '\n\n']);

                fprintf(['Generation of impulse response matrix, forward method (CPU time): '...
                    num2str(t_fwd(rt_iter,avg_iter)) ' sec \n']);

                fprintf(['Generation of impulse response matrix, adjoint method (CPU time): '...
                    num2str(t_adj(rt_iter,avg_iter)) ' sec \n']);

                fprintf(['Generation of impulse response matrix, ON3 (CPU time): '...
                    num2str(t_ON3(rt_iter,avg_iter)) ' sec \n\n']);

                fprintf(['Comparison of impulse response matrices, max(max(Xi_fwd - Xi_adj)): '...
                    num2str(max(max(Xi_fwd-Xi_adj))) '\n']);

                fprintf(['Comparison of impulse response matrices, max(max(Xi_fwd - Xi_ON3)): '...
                    num2str(max(max(Xi_fwd-Xi_ON3))) '\n']);

                fprintf(['Comparison of impulse response matrices, max(max(Xi_ON3 - Xi_adj)): '...
                    num2str(max(max(Xi_ON3-Xi_adj))) '\n']);

                % print matrices to screen if they are small enough
                if N < 6
                    fprintf('Impulse response matrix, forward method:\n')
                    Xi_fwd
                    fprintf('Impulse response matrix, adjoint method:\n')
                    Xi_adj
                    fprintf('Impulse response matrix, ON3:\n')
                    Xi_ON3
                    fprintf('\n')
                end
            end
    end
end
if run.plot_runtime_fig == true

    % calculate average of all iterations
    t_cvodes_avg = zeros(length(Nx),1);
    t_fwd_avg = zeros(length(Nx),1);
    t_adj_avg = zeros(length(Nx),1);
    t_ON3_avg = zeros(length(Nx),1);
    for k = 1:avg_pts
        t_cvodes_avg = t_cvodes_avg + t_cvodes(:,k);
        t_fwd_avg = t_fwd_avg + t_fwd(:,k);
        t_adj_avg = t_adj_avg + t_adj(:,k);
        t_ON3_avg = t_ON3_avg + t_ON3(:,k);
    end
    t_cvodes_avg = t_cvodes_avg/avg_pts;
    t_fwd_avg = t_fwd_avg/avg_pts;
    t_adj_avg = t_adj_avg/avg_pts;
    t_ON3_avg = t_ON3_avg/avg_pts;

    %% RUNTIME FIGURES

    switch run.rt_mode
        case 1
            figure()
            plot(rt_Nx,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('N_x');
            ylabel('CPU time [s]');
            x_lim = xlim();
            %axis([x_lim(1) x_lim(2) 0 t_axis_c]);
            grid;

            figure()
            plot(rt_Nx,t_fwd_avg,'-*');
            hold all;
            plot(rt_Nx,t_adj_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('Forward','Adjoint');
            xlabel('N');
            ylabel('CPU time [s]');
            x_lim = xlim();
            %axis([x_lim(1) x_lim(2) 0 t_axis_fa]);
            grid;

            figure()
            plot(rt_Nx,t_ON3_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            x_lim = xlim();
            %axis([x_lim(1) x_lim(2) 0 t_axis_on3]);
            grid;
        case 2
            % does not apply to this mode
        otherwise
            fprintf('ERROR: rt_mode not valid');
    end
end
end

%% EOF