function cvodes_pde_alt
%% Main function for CVODES sensitivity integration (alternative method)
%% and generation of impulse response matrix for heat equation (PDE)

clear all
close all
clc
format compact
format short

%% SUNDIALS STARTUP

w = pwd;
sundials_path = '/Applications/MATLAB_R2008a/toolbox/sundialsTB';
cd(sundials_path);
startup_STB();
cd(w);

%% SET RUN OPTIONS

run = struct;
run.plot_fig         = 1; % plot solution
run.plot_fig_title   = 1; % title names in figures
run.plot_runtime_fig = 1; % plot runtime figurfalses
run.print_info       = 1; % print information
run.rt_mode          = 2; % choose runtime measurement mode

mondata = struct;       % monitor data struct
mondata.stats = 0;      % report step size and method order
mondata.cntr  = 0;      % report counter evolution
mondata.mode  = 'both'; % graphical/text/both
mondata.sol   = 0;      % plot solution
mondata.sensi = 0;      % plot sensitivities (FSA)

% different modes for measuring runtime
switch run.rt_mode
    case 1
        rt_Nx = 2:6:200;
        pts = 10;
        avg_pts = 10; % number of iterations to average over
    case 2
        rt_Nx = 50;
        pts = 50;
        avg_pts = 1; % number of iterations to average over
    otherwise
        fprintf('ERROR: rt_mode not valid');
end

%% USER DATA / PARAMETERS

global ud;  % user data structure

% time
t0 = 0;
tf = 0.5;
Ts = tf/pts;
t = t0:Ts:tf;
N = length(t);

for avg_iter = 1:avg_pts;
    for rt_iter = 1:length(rt_Nx);

        % system
        Nx = rt_Nx(rt_iter); %must be even
        Nu = 1;
        Nz = Nx; % full state measurement
        states = 1/Nx:1/Nx:1;
        delta = 1/Nx;
        v = [-2 1 zeros(1,Nx-2)];
        ud.A = toeplitz(v,v)/delta^2;
        ud.B = [1 zeros(1,Nx-1)]'/delta^2;

        fprintf(['Iteration ' num2str(rt_iter) '/' num2str(length(rt_Nx)) ...
            ', Average Iteration ' num2str(avg_iter) '/' num2str(avg_pts) '\n']);

        %% CVODES INIT

            options = CVodeSetOptions('RelTol',1.e-6,...
                'AbsTol',1.e-6,...
                'JacobianFn',@cvodes_pde_J);

            options = CVodeSetOptions(options,'MonitorFn',@CVodeMonitor,...
                'MonitorData',mondata,...
                'MaxStep',Ts);

            x0 = zeros(Nx,1);                         % initial values
            CVodeMalloc(@cvodes_pde_f,t0,x0,options); % allocate memory

            %% FORWARD SENSITIVITY ANALYSIS INIT

            FSAoptions = CVodeSetFSAOptions('SensErrControl', 'on',...
                'SensRhsFn', @cvodes_pde_alt_fS);

            xS0 = zeros(Nx,Nu);                                    % sensitivity initial values
            CVodeSensMalloc(Nu, 'Simultaneous', xS0, FSAoptions);  % allocate memory

            %% PROBLEM SOLUTION

            % prealloc for speed
            x = zeros(Nx,N);
            S = zeros(Nx,Nu,(N-1),(N-1));

            % sinus input
            periods = 2;
            freq = periods/tf;
            omega = 2*pi*freq;
            u = sin(omega*t);

            x(:,1) = x0;
            tic;
            for k = 1:N-1

                ud.sens_Ju = true;

                % update control
                ud.u = u(k);

                % reinit system equations
                CVodeReInit(@cvodes_pde_f,t(k),x(:,k),options);

                % reinit sensitivity equations
                CVodeSensReInit('Simultaneous', xS0, FSAoptions);

                % integration
                [status, tk, x(:,k+1), S(:,:,k,k)] = CVode(t(k+1),'Normal');

                ud.sens_Ju = false;

                for n = k+1:N-1

                    % update control
                    ud.u = u(n);

                    % reinit system equations
                    CVodeReInit(@cvodes_pde_f,t(n),x(:,n),options);

                    % reinit sensitivity equations
                    CVodeSensReInit('Simultaneous', S(:,:,n-1,k), FSAoptions);

                    % integration
                    [status, tk, x(:,n+1), S(:,:,n,k)] = CVode(t(n+1),'Normal');

                end
            end
            t_cvodes(rt_iter,avg_iter) = toc;

            si = CVodeGetStats; % store stats

            CVodeFree;  % free memory

            if run.plot_fig == true

                %% FIGURES

                figure()
                mesh(states,t,x')
                if run.plot_fig_title == true
                    title(['solution']);
                end
                axis([0 1 0 tf -1 1])
                title('Zero state response - sinusoidal input')
                xlabel('x')
                ylabel('t')
                zlabel('temperature')

                figure();
                plot(t,u','-*');
                if run.plot_fig_title == true
                    title(['control']);
                end
                legend('u');
                xlabel('t [s]');
                grid;
            end

            C = zeros(Nx,Nx,N);
            D = zeros(Nz,Nu,N);
            for k = 1:N
                C(:,:,k) = eye(Nx);
            end

            %% BUILD IMPULSE RESPONSE MATRIX - ALTERNATIVE METHOD

            % prealloc for speed
            Xi_alt = zeros(N*Nz,N*Nu);

            tic;
            for k = 1:N
                Xi_alt((k-1)*Nz+1:k*Nz,(k-1)*Nu+1:k*Nu) = D(:,:,k);
                for n = k+1:N
                    Xi_alt((n-1)*Nz+1:n*Nz,(k-1)*Nu+1:k*Nu) = C(:,:,k)*S(:,:,n-1,k);
                end
            end
            t_alt(rt_iter,avg_iter) = toc;

            if run.print_info == true

                %% PRINT INFORMATION

                fprintf(['INFORMATION\n']);

                fprintf(['Start time: ' num2str(t0) '\n']);
                fprintf(['Sampling time: ' num2str(Ts) '\n']);
                fprintf(['Stop time: ' num2str(tf) '\n']);
                fprintf(['Number of sampling points: ' num2str(N) '\n\n']);

                fprintf(['CVODES integration time: ' num2str(t_cvodes(rt_iter,avg_iter)) '\n\n']);

                fprintf(['Generation of impulse response matrix, (CPU time): '...
                    num2str(t_alt(rt_iter,avg_iter)) ' sec \n\n']);

                % print matrices to screen if they are small enough
                if N < 6
                    fprintf('Impulse response matrix, alternative method:\n')
                    Xi_alt
                    fprintf('\n')
                end
            end
    end
end
if run.plot_runtime_fig == true

    % calculate average of all iterations
    t_cvodes_avg = zeros(length(Nx),1);
    t_alt_avg = zeros(length(Nx),1);
    for k = 1:avg_pts
        t_cvodes_avg = t_cvodes_avg + t_cvodes(:,k);
        t_alt_avg = t_alt_avg + t_alt(:,k);
    end
    t_cvodes_avg = t_cvodes_avg/avg_pts;
    t_alt_avg = t_alt_avg/avg_pts;

    %% RUNTIME FIGURES

    switch run.rt_mode
        case 1
            figure()
            plot(rt_Nx,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('N_x');
            ylabel('CPU time [s]');
            x_lim = xlim();
            %axis([x_lim(1) x_lim(2) 0 t_axis_c]);
            grid;

            figure()
            plot(rt_Nx,t_alt_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('Alternative');
            xlabel('N_x');
            ylabel('CPU time [s]');
            x_lim = xlim();
            %axis([x_lim(1) x_lim(2) 0 t_axis_alt]);
            grid;
        case 2
            % does not apply to this mode
        otherwise
            fprintf('ERROR: rt_mode not valid');
    end
end
end

%% EOF