function cvodes_vdv
%% Main function for CVODES sensitivity computation
%% and generation of impulse response matrix for Van de Vusse reactor

clear all
close all
clc
format compact
format short

%% SUNDIALS STARTUP

w = pwd;
sundials_path = '/Applications/MATLAB_R2008a/toolbox/sundialsTB';
cd(sundials_path);
startup_STB();
cd(w);

%% SET OPTIONS (true = 1, false = 0)

run = struct;
run.plot_fig         = 1; % plot solution
run.plot_fig_title   = 1; % title names in figures
run.plot_runtime_fig = 1; % plot runtime figures
run.print_info       = 1; % print information
run.rt_mode          = 4; % choose runtime measurement mode

mondata = struct;        % monitor data struct
mondata.mon_en = 0;      % enable monitoring
mondata.stats  = 0;      % report step size and method order
mondata.cntr   = 0;      % report counter evolution
mondata.mode   = 'both'; % graphical/text/both
mondata.sol    = 0;      % plot solution
mondata.sensi  = 0;      % plot sensitivities (FSA)

pts = 20;    % number of points to measure runtime
avg_pts = 100; % number of iterations to average over

% different modes for measuring runtime
switch run.rt_mode
    case 1
        % tf constant, Ts variable, N variable
        rt_N  = (100/pts:100/pts:100);
        rt_t0 = zeros(1,pts);
        rt_tf = ones(1,pts);
        rt_Ts = (rt_tf - rt_t0)./rt_N;
    case 2
        % tf variable, Ts variable, N constant
        rt_N  = 100*ones(1,pts);
        rt_tf = 1/pts:1/pts:1;
        rt_t0 = zeros(1,pts);
        rt_Ts = (rt_tf - rt_t0)./rt_N;
    case 3
        % tf variable, Ts constant, N variable
        rt_Ts = 1/100*ones(1,pts);
        rt_t0 = zeros(1,pts);
        rt_tf = 1/pts:1/pts:1;
        rt_N  = (rt_tf-rt_t0)./rt_Ts;
    case 4
        rt_t0 = 0;
        rt_Ts = 1/4;
        rt_tf = 1;
        rt_N  = (rt_tf-rt_t0)./rt_Ts;
        pts   = 1;
        avg_pts = 1;
    otherwise
        fprintf('ERROR: rt_mode not valid');
end

for k = 1:pts
    rt_t{k} = (rt_t0:rt_Ts(k):rt_tf(k));
end

%% USER DATA / PARAMETERS

global ud;  % user data structure

% dimensions
Nu = 2;     % dimension of u
Nx = 4;     % dimension of x
Nz = 4;     % dimension of z

% objective function parameters - sum(xk'*Q*xk + uk*'R*uk)/2 + xN'*P*xN/2
P = eye(Nx);
Q = eye(Nz);
R = eye(Nu);

% parameters
ud.alpha0   = 30.8285;
ud.gamma0  = 0.1;
ud.k10     = 1.287e12;
ud.k20     = 9.043e6;
ud.DHab    = 4.2;
ud.DHad    = -41.85;
ud.cin     = 5100;
ud.beta0   = 86.688;
ud.delta0   = 3.556e-4;
ud.E1      = 9758.3;
ud.E2      = 8560;
ud.DHbc    = -11;
ud.Tin     = 104.9;

for avg_iter = 1:avg_pts;
    for rt_iter = 1:length(rt_t);

        % time
        t0 = rt_t0(rt_iter);
        tf = rt_tf(rt_iter);
        Ts = rt_Ts(rt_iter);
        t = rt_t{rt_iter};
        N = length(t);

        x0 = [0 750 85 100]';                      % initial values

        % step inputs
        u1 = [   12*ones(1,floor(2*N/3))  1*ones(1,N-floor(2*N/3))];
        u2 = [-3000*ones(1,floor(N/3)) -1000*ones(1,N-floor(N/3))];

        u = [u1; u2];

        fprintf(['Iteration ' num2str(rt_iter) '/' num2str(pts) ...
            ', Average Iteration ' num2str(avg_iter) '/' num2str(avg_pts) '\n']);

        %% SET CVODES OPTIONS

            options = CVodeSetOptions('RelTol',1.e-6,'AbsTol',1.e-6,...
                'JacobianFn',@cvodes_vdv_J,'MaxStep',Ts);

            if mondata.mon_en == true
                options = CVodeSetOptions(options,'MonitorFn',@CVodeMonitor,...
                    'MonitorData',mondata);
            end

            %% CVODES INIT

            CVodeMalloc(@cvodes_vdv_f,t0,x0,options); % allocate memory

            % FSA init

            FSAoptions = CVodeSetFSAOptions('SensErrControl', 'on',...
                'SensRhsFn', @cvodes_vdv_fS);

            xS0 = [eye(Nx) zeros(Nx,Nu)];                               % sensitivity initial values
            CVodeSensMalloc(Nx + Nu, 'Simultaneous', xS0, FSAoptions);  % allocate memory

            %% PROBLEM SOLUTION

            % prealloc for speed
            x  = zeros(Nx,N);
            AS = zeros(Nx,Nx,N-1);
            BS = zeros(Nx,Nu,N-1);

            x(:,1) = x0;
            tic;
            for k = 2:N
                % update control
                ud.u1 = u1(k-1);
                ud.u2 = u2(k-1);

                % reinit system equations
                CVodeReInit(@cvodes_vdv_f,t(k)-Ts,x(:,k-1),options);

                % reinit sensitivity equations
                CVodeSensReInit('Simultaneous', xS0, FSAoptions);

                % integration
                [status, tk, xk, xSk] = CVode(t(k),'Normal');

                % store data
                x(:,k) = xk;
                AS(:,:,k) = xSk(:,1:Nx);
                BS(:,:,k) = xSk(:,Nx+1:Nx+Nu);
            end
            t_cvodes(rt_iter,avg_iter) = toc;

            si = CVodeGetStats; % store stats

            CVodeFree;  % free memory

            if run.plot_fig == true

                %% FIGURES

                figure();
                plot(t,x(1:2,:)','-*');
                if run.plot_fig_title == true
                    title(['solution']);
                end
                legend('c_A','c_B');
                xlabel('t [h]');
                ylabel('Concentration [mol/m^3]');
                grid;

                figure();
                plot(t,x(3:4,:)','-*');
                if run.plot_fig_title == true
                    title(['solution']);
                end
                legend('T','T_c');
                xlabel('t [h]');
                ylabel('Temperature [C]');
                grid;

                figure();
                plot(t,u1','-*');
                hold all
                plot(t,u2','-*');
                if run.plot_fig_title == true
                    title(['control']);
                end
                legend('u_1','u_2');
                xlabel('t [h]');
                grid;

                state_names = cell(Nx,1);
                state_names{1} = 'c_A';
                state_names{2} = 'c_B';
                state_names{3} = 'T';
                state_names{4} = 'T_c';

                control_names = cell(Nx,1);
                control_names{1} = 'u_1';
                control_names{2} = 'u_2';

                % plot sensitivities
                for k = 1:Nx
                    figure();
                    for n = 1:Nx
                        AS_plot = zeros(1,N);
                        for i = 1:N
                            AS_plot(i) = AS(n,k,i);
                        end
                        plot(t,AS_plot','-*');
                        hold all;
                    end
                    if run.plot_fig_title == true
                        title([state_names{k} ' sensitivities']);
                    end
                    legend('c_A','c_B','T','T_c');
                    xlabel('t [h]');
                    ylabel('Sensitivity');
                    grid;

                end
                for k = 1:Nu
                    figure();
                    for n = 1:Nx
                        BS_plot = zeros(1,N);
                        for i = 1:N
                            BS_plot(i) = BS(n,k,i);
                        end
                        plot(t,BS_plot','-*');
                        hold all;
                    end
                    if run.plot_fig_title == true
                        title([control_names{k} ' sensitivities']);
                    end
                    legend('c_A','c_B','T','T_c');
                    xlabel('t [h]');
                    ylabel('Sensitivity');
                    grid;

                end
            end

            % define matrices X = [X_1 X_2 ... X_N]
            A = AS(:,:,2:N);
            B = BS(:,:,2:N);

            % choose wheter to have full state measurement
            if Nz == 4
                C = zeros(Nz,Nx,N);
                D = zeros(Nz,Nu,N);
                for k = 1:N
                    C(:,:,k) = eye(4);
                end
            elseif Nz == 2
                C = zeros(Nz,Nx,N);
                D = zeros(Nz,Nu,N);
                for k = 1:N
                    C(:,:,k) = [0 1 0 0 ; 0 0 1 0];
                    D(:,:,k) = eye(2);
                end
            else
                fprintf('ERROR: C not defined for this number of outputs\n');
            end

            %% BUILD IMPULSE RESPONSE MATRIX - FORWARD METHOD

            % prealloc for speed
            Xi_fwd = zeros(N*Nz,N*Nu);
            phi = zeros(Nx,Nu*(N-1));

            tic;
            Xi_fwd(1:Nz,1:Nu) = D(:,:,1);
            for j=2:N
                for i=1:j-2
                    phi(1:Nx,(i-1)*Nu+1:i*Nu) = A(:,:,j-1)*phi(1:Nx,(i-1)*Nu+1:i*Nu);
                end
                phi(1:Nx,(j-2)*Nu+1:(j-1)*Nu) = B(:,:,j-1);
                for i = 1:j-1
                    Xi_fwd((j-1)*Nz+1:j*Nz,(i-1)*Nu+1:i*Nu) = C(:,:,j)*phi(1:Nx,(i-1)*Nu+1:i*Nu);
                end
                Xi_fwd((j-1)*Nz+1:j*Nz,(j-1)*Nu+1:j*Nu) = D(:,:,j);
            end
            t_imp_fwd(rt_iter,avg_iter) = toc;


            %% BUILD IMPULSE RESPONSE MATRIX - ADJOINT METHOD

            % prealloc for speed
            Xi_adj = zeros(N*Nz,N*Nu);
            lambda = zeros(Nz,Nx*(N-1));

            tic;
            Xi_adj((N-1)*Nz+1:N*Nz,(N-1)*Nu+1:N*Nu) = D(:,:,N);
            for k = N-1:-1:1
                lambda((k-1)*Nz+1:k*Nz,1:Nx) = C(:,:,k+1);
                for i = k+1:N-1
                    lambda((i-1)*Nz+1:i*Nz,1:Nx) = lambda((i-1)*Nz+1:i*Nz,1:Nx)*A(:,:,k+1);
                end
                Xi_adj((k-1)*Nz+1:k*Nz,(k-1)*Nu+1:k*Nu) = D(:,:,k);
                for i = k:N-1
                    Xi_adj(i*Nz+1:(i+1)*Nz,(k-1)*Nu+1:k*Nu) = lambda((i-1)*Nz+1:i*Nz,1:Nx)*B(:,:,k);
                end
            end
            t_imp_adj(rt_iter,avg_iter) = toc;


            %% BUILD IMPULSE RESPONSE MATRIX - ON3

            % prealloc for speed
            Theta_ON3 = zeros(N*Nx,N*Nu);
            Xi_ON3    = zeros(N*Nz,N*Nu);

            tic;
            for k=1:N-1
                for j=k+1:N
                    Theta_ON3((j-1)*Nx+1:j*Nx,(k-1)*Nu+1:k*Nu) = B(:,:,k);
                end

                for i=1:k
                    for j = k+2:N
                        Theta_ON3((j-1)*Nx+1:j*Nx,(i-1)*Nu+1:i*Nu) = ...
                            A(:,:,k+1)*Theta_ON3((j-1)*Nx+1:j*Nx,(i-1)*Nu+1:i*Nu);
                    end
                end
            end

            for k=1:N
                Ck = C(:,:,k);
                Dk = D(:,:,k);
                for i=1:k-1
                    Xi_ON3((k-1)*Nz+1:k*Nz,(i-1)*Nu+1:i*Nu) = ...
                        Ck*Theta_ON3((k-1)*Nx+1:k*Nx,(i-1)*Nu+1:i*Nu);
                end
                Xi_ON3((k-1)*Nz+1:k*Nz,(k-1)*Nu+1:k*Nu) = Dk;
            end
            t_imp_ON3(rt_iter,avg_iter) = toc;


            if run.print_info == true

                %% PRINT INFORMATION

                fprintf(['INFORMATION\n']);

                fprintf(['Start time: ' num2str(t0) '\n']);
                fprintf(['Sampling time: ' num2str(Ts) '\n']);
                fprintf(['Stop time: ' num2str(tf) '\n']);
                fprintf(['Number of sampling points: ' num2str(N) '\n\n']);

                fprintf(['CVODES integration time: ' num2str(t_cvodes(rt_iter,avg_iter)) '\n\n']);

                fprintf(['Generation of impulse response matrix, forward method (CPU time): '...
                    num2str(t_imp_fwd(rt_iter,avg_iter)) ' sec \n']);

                fprintf(['Generation of impulse response matrix, adjoint method (CPU time): '...
                    num2str(t_imp_adj(rt_iter,avg_iter)) ' sec \n']);

                fprintf(['Generation of impulse response matrix, ON3 (CPU time): '...
                    num2str(t_imp_ON3(rt_iter,avg_iter)) ' sec \n\n']);

                fprintf(['Comparison of impulse response matrices, max(max(Xi_fwd - Xi_adj)): '...
                    num2str(max(max(Xi_fwd-Xi_adj))) '\n']);

                fprintf(['Comparison of impulse response matrices, max(max(Xi_fwd - Xi_ON3)): '...
                    num2str(max(max(Xi_fwd-Xi_ON3))) '\n']);

                fprintf(['Comparison of impulse response matrices, max(max(Xi_ON3 - Xi_adj)): '...
                    num2str(max(max(Xi_ON3-Xi_adj))) '\n']);

                % print matrices to screen if they are small enough
                if N < 6
                    fprintf('Impulse response matrix, forward method:\n')
                    Xi_fwd
                    fprintf('Impulse response matrix, adjoint method:\n')
                    Xi_adj
                    fprintf('Impulse response matrix, ON3:\n')
                    Xi_ON3
                    fprintf('\n')
                end
            end
    end
end
if run.plot_runtime_fig == true

    % calculate average of all iterations
    t_cvodes_avg = zeros(pts,1);
    t_imp_fwd_avg = zeros(pts,1);
    t_imp_adj_avg = zeros(pts,1);
    t_imp_ON3_avg = zeros(pts,1);
    for k = 1:avg_pts
        t_cvodes_avg = t_cvodes_avg + t_cvodes(:,k);
        t_imp_fwd_avg = t_imp_fwd_avg + t_imp_fwd(:,k);
        t_imp_adj_avg = t_imp_adj_avg + t_imp_adj(:,k);
        t_imp_ON3_avg = t_imp_ON3_avg + t_imp_ON3(:,k);
    end
    t_cvodes_avg = t_cvodes_avg/avg_pts;
    t_imp_fwd_avg = t_imp_fwd_avg/avg_pts;
    t_imp_adj_avg = t_imp_adj_avg/avg_pts;
    t_imp_ON3_avg = t_imp_ON3_avg/avg_pts;

    %% RUNTIME FIGURES

    switch run.rt_mode
        case 1
            % tf constant, Ts variable, N variable
            figure()
            plot(rt_N,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_N,t_imp_fwd_avg,'-*');
            hold all;
            plot(rt_N,t_imp_adj_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('Forward','Adjoint');
            xlabel('N');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_N,t_imp_ON3_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;
        case 2
            % tf variable, Ts variable, N constant
            figure()
            plot(rt_tf,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('t_f');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_tf,t_imp_fwd_avg,'-*');
            hold all;
            plot(rt_tf,t_imp_adj_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('Forward','Adjoint');
            xlabel('t_f');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_tf,t_imp_ON3_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('O(N^3)');
            xlabel('t_f');
            ylabel('CPU time [s]');
            grid;
        case 3
            % tf variable, Ts constant, N variable
            figure()
            plot(rt_N,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_N,t_imp_fwd_avg,'-*');
            hold all;
            plot(rt_N,t_imp_adj_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('Forward','Adjoint');
            xlabel('N');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_N,t_imp_ON3_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            legend('O(N^3)');
            xlabel('N');
            ylabel('CPU time [s]');
            grid;
        case 4
            % does not apply to this mode
        otherwise
            fprintf('ERROR: rt_mode not valid');
    end
end
end

%% EOF