function cvodes_vdv_alt
%% Main function for CVODES sensitivity integration (alternative method)
%% and generation of impulse response matrix for Van de Vusse reactor

clear all
close all
clc
format compact
format short

%% SUNDIALS STARTUP

w = pwd;
sundials_path = '/Applications/MATLAB_R2008a/toolbox/sundialsTB';
cd(sundials_path);
startup_STB();
cd(w);

%% SET RUN OPTIONS

run = struct;
run.plot_fig         = 1; % plot solution
run.plot_fig_title   = 1; % title names in figures
run.plot_runtime_fig = 1; % plot runtime figures
run.print_info       = 1; % print information
run.rt_mode          = 4; % choose runtime measurement mode

mondata = struct;       % monitor data struct
mondata.stats = 0;      % report step size and method order
mondata.cntr  = 0;      % report counter evolution
mondata.mode  = 'both'; % graphical/text/both
mondata.sol   = 0;      % plot solution
mondata.sensi = 0;      % plot sensitivities (FSA)

pts = 10;    % number of points to measure runtime
avg_pts = 50; % number of iterations to average over

% different modes for measuring runtime
switch run.rt_mode
    case 1
        % tf constant, Ts variable, N variable
        rt_N  = 1:100/pts:100;
        rt_t0 = zeros(1,pts);
        rt_tf = ones(1,pts);
        rt_Ts = (rt_tf - rt_t0)./rt_N;
    case 2
        % tf variable, Ts variable, N constant
        rt_N  = 100*ones(1,pts);
        rt_tf = 1/pts:1/pts:1;
        rt_t0 = zeros(1,pts);
        rt_Ts = (rt_tf - rt_t0)./rt_N;
    case 3
        % tf variable, Ts constant, N variable
        rt_Ts = 1/100*ones(1,pts);
        rt_t0 = zeros(1,pts);
        rt_tf = 1/pts:1/pts:1;
        rt_N  = (rt_tf-rt_t0)./rt_Ts;
    case 4
        rt_t0 = 0;
        rt_Ts = 1/4;
        rt_tf = 1;
        rt_N  = (rt_tf-rt_t0)./rt_Ts;
        pts   = 1;
        avg_pts = 1;
    otherwise
        fprintf('ERROR: rt_mode not valid');
end

for k = 1:pts
    rt_t{k} = (rt_t0:rt_Ts(k):rt_tf(k));
end

%% USER DATA / PARAMETERS

global ud;  % user data structure

% dimensions
Nu = 2;     % dimension of u
Nx = 4;     % dimension of x
Nz = 4;     % dimension of z

% parameters
ud.alpha0   = 30.8285;
ud.gamma0  = 0.1;
ud.k10     = 1.287e12;
ud.k20     = 9.043e6;
ud.DHab    = 4.2;
ud.DHad    = -41.85;
ud.cin     = 5100;
ud.beta0   = 86.688;
ud.delta0   = 3.556e-4;
ud.E1      = 9758.3;
ud.E2      = 8560;
ud.DHbc    = -11;
ud.Tin     = 104.9;

for avg_iter = 1:avg_pts;
    for rt_iter = 1:length(rt_t);

        % time
        t0 = rt_t0(rt_iter);
        tf = rt_tf(rt_iter);
        Ts = rt_Ts(rt_iter);
        t = rt_t{rt_iter};
        N = length(t);

        fprintf(['Iteration ' num2str(rt_iter) '/' num2str(pts) ...
            ', Average Iteration ' num2str(avg_iter) '/' num2str(avg_pts) '\n']);

        %% CVODES INIT

            options = CVodeSetOptions('RelTol',1.e-6,...
                'AbsTol',1.e-6,...
                'JacobianFn',@cvodes_vdv_J);

            options = CVodeSetOptions(options,'MonitorFn',@CVodeMonitor,...
                'MonitorData',mondata,...
                'MaxStep',Ts);

            x0 = [0 750 85 100]';                      % initial values
            CVodeMalloc(@cvodes_vdv_f,t0,x0,options); % allocate memory

            %% FORWARD SENSITIVITY ANALYSIS INIT

            FSAoptions = CVodeSetFSAOptions('SensErrControl', 'on',...
                'SensRhsFn', @cvodes_vdv_alt_fS);

            xS0 = zeros(Nx,Nu);                          % sensitivity initial values
            CVodeSensMalloc(Nu, 'Simultaneous', xS0, FSAoptions);  % allocate memory

            %% PROBLEM SOLUTION

            % prealloc for speed
            x = zeros(Nx,N);
            S = zeros(Nx,Nu,(N-1),(N-1));


            % step inputs
            u1 = [   12*ones(1,floor(2*N/3))  1*ones(1,N-floor(2*N/3))];
            u2 = [-3000*ones(1,floor(N/3)) -1000*ones(1,N-floor(N/3))];

            x(:,1) = x0;
            tic;
            for k = 1:N-1

                ud.sens_Ju = true;

                % update control
                ud.u1 = u1(k);
                ud.u2 = u2(k);

                % reinit system equations
                CVodeReInit(@cvodes_vdv_f,t(k),x(:,k),options);

                % reinit sensitivity equations
                CVodeSensReInit('Simultaneous', xS0, FSAoptions);

                % integration
                [status, tk, x(:,k+1), S(:,:,k,k)] = CVode(t(k+1),'Normal');

                ud.sens_Ju = false;

                for n = k+1:N-1

                    % update control
                    ud.u1 = u1(n);
                    ud.u2 = u2(n);

                    % reinit system equations
                    CVodeReInit(@cvodes_vdv_f,t(n),x(:,n),options);

                    % reinit sensitivity equations
                    CVodeSensReInit('Simultaneous', S(:,:,n-1,k), FSAoptions);

                    % integration
                    [status, tk, x(:,n+1), S(:,:,n,k)] = CVode(t(n+1),'Normal');

                end
            end
            t_cvodes(rt_iter,avg_iter) = toc;

            si = CVodeGetStats; % store stats

            CVodeFree;  % free memory

            if run.plot_fig == true

                %% FIGURES

                figure();
                plot(t,x(1:2,:)','-*');
                if run.plot_fig_title == true
                    title(['solution']);
                end
                legend('c_A','c_B');
                xlabel('t [s]');
                ylabel('Concentration [mol/m^3]');
                grid;

                figure();
                plot(t,x(3:4,:)','-*');
                if run.plot_fig_title == true
                    title(['solution']);
                end
                legend('T','T_c');
                xlabel('t [s]');
                ylabel('Temperature [C]');
                grid;

                figure();
                plot(t,u1','-*');
                hold all
                plot(t,u2','-*');
                if run.plot_fig_title == true
                    title(['control']);
                end
                legend('u_1','u_2');
                xlabel('t [s]');
                grid;
            end

            % choose wheter to have full state measurement
            if Nz == 4
                C = zeros(Nx,Nx,N);
                D = zeros(Nz,Nu,N);
                for k = 1:N
                    C(:,:,k) = eye(4);
                end
            elseif Nz == 2
                C = zeros(Nz,Nx,N);
                D = zeros(Nz,Nu,N);
                for k = 1:N
                    C(:,:,k) = [0 1 0 0 ; 0 0 1 0];
                end
            else
                fprintf('ERROR: C not defined for this number of outputs\n');
            end

            %% BUILD IMPULSE RESPONSE MATRIX

            % prealloc for speed
            Xi_alt = zeros(N*Nz,N*Nu);

            tic;
            for k = 1:N
                Xi_alt((k-1)*Nz+1:k*Nz,(k-1)*Nu+1:k*Nu) = D(:,:,k);
                for n = k+1:N
                    Xi_alt((n-1)*Nz+1:n*Nz,(k-1)*Nu+1:k*Nu) = C(:,:,n)*S(:,:,n-1,k);
                end
            end
            t_alt(rt_iter,avg_iter) = toc;

            if run.print_info == true

                %% PRINT INFORMATION

                fprintf(['INFORMATION\n']);

                fprintf(['Start time: ' num2str(t0) '\n']);
                fprintf(['Sampling time: ' num2str(Ts) '\n']);
                fprintf(['Stop time: ' num2str(tf) '\n']);
                fprintf(['Number of sampling points: ' num2str(N) '\n\n']);

                fprintf(['CVODES integration time: ' num2str(t_cvodes(rt_iter,avg_iter)) '\n\n']);

                fprintf(['Generation of impulse response matrix, (CPU time): '...
                    num2str(t_alt(rt_iter,avg_iter)) ' sec \n\n']);

                % print matrices to screen if they are small enough
                if N < 6
                    fprintf('Impulse response matrix, alternative method:\n')
                    Xi_alt
                    fprintf('\n')
                end
            end
    end
end
if run.plot_runtime_fig == true

    % calculate average of all iterations
    t_cvodes_avg = zeros(pts,1);
    t_alt_avg = zeros(pts,1);
    for k = 1:avg_pts
        t_cvodes_avg = t_cvodes_avg + t_cvodes(:,k);
        t_alt_avg = t_alt_avg + t_alt(:,k);
    end
    t_cvodes_avg = t_cvodes_avg/avg_pts;
    t_alt_avg = t_alt_avg/avg_pts;

    switch run.rt_mode
        case 1
            % tf constant, Ts variable, N variable
            figure()
            plot(rt_N,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_N,t_alt_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;
        case 2
            % tf variable, Ts variable, N constant
            figure()
            plot(rt_tf,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('t_f');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_tf,t_alt_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            xlabel('t_f');
            ylabel('CPU time [s]');
            grid;
        case 3
            figure()
            plot(rt_N,t_cvodes_avg,'-*');
            if run.plot_fig_title == true
                title(['CVODES integration runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;

            figure()
            plot(rt_N,t_alt_avg,'-*');
            hold all;
            if run.plot_fig_title == true
                title(['Build impulse response matrix runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;
        case 4
            % does not apply to this mode
        otherwise
            fprintf('ERROR: rt_mode not valid');
    end
end
end

%% EOF