function finite_diff_vdv()
%% Calculates sensitivities using finite differences

clear all
close all
clc
format compact
format short

% dimensions
Nu = 2;     % dimension of u
Nx = 4;     % dimension of x
Nz = 4;     % dimension of z

%% reactor parameters
global ud;
ud.alpha0   = 30.8285;
ud.gamma0  = 0.1;
ud.k10     = 1.287e12;
ud.k20     = 9.043e6;
ud.DHab    = 4.2;
ud.DHad    = -41.85;
ud.cin     = 5100;
ud.beta0   = 86.688;
ud.delta0   = 3.556e-4;
ud.E1      = 9758.3;
ud.E2      = 8560;
ud.DHbc    = -11;
ud.Tin     = 104.9;

x0 = [0; 750; 85; 100];

Ts = 1/4;
t0 = 0;
tf = 1;

t = (t0:Ts:tf); % time vector
N = length(t);  % horizon length

% step inputs
u1 = [   12*ones(1,floor(2*N/3))  1*ones(1,N-floor(2*N/3))];
u2 = [-3000*ones(1,floor(N/3)) -1000*ones(1,N-floor(N/3))];

u = [u1; u2];
x = vdv_sim(t,u,x0,1,N);

figure()
plot(t,x(1:2,:))
legend('c_A','c_B');
grid

figure()
plot(t,x(3:4,:))
legend('T','T_c');
grid

%% BUILD IMPULSE RESPONE MATRIX - FINITE DIFF
tic;
eps = 1e-6;
Xi_fin = zeros(N*Nz,N*Nu);
u = [u1 ; u2];
for k = 1:N-1 % column
    for n = 1:Nu
        u_pert = u;
        u_pert(n,k) = u_pert(n,k)+eps;
        x_pert = vdv_sim(t,u_pert,x(:,1:k),k,N);
        Xi_fin(:,(k-1)*Nu+n) = mat2vec((x_pert-x)/eps,Nx,N);
    end
end
toc
Xi_fin
end


function x_ret = vdv_sim(t,u,xk,k_start,k_stop)

global ud

options = odeset('RelTol',1.e-8,'AbsTol',1.e-8);

x_ret(:,1:k_start) = xk;
x0 = xk(:,k_start)';
N = length(t);
for n = k_start+1:k_stop
    ud.u1 = u(1,n-1);
    ud.u2 = u(2,n-1);
    [t_sol,x_sol] = ode45(@van_de_vusse,[t(n-1) t(n)],x0,options);
    x0 = x_sol(length(x_sol),:);
    x_ret(:,n) = x_sol(length(x_sol),:)';
end

end

function dx = van_de_vusse(t,y)
    global ud
    
    dx = zeros(4,1);
    dx(1) = -ud.k10*exp(-ud.E1/(y(3)+273.15))*y(1)...
            - ud.k20*exp(-ud.E2/(y(3)+273.15))*y(1)^2 ...
            + (ud.cin-y(1))*ud.u1;
        
    dx(2) = ud.k10*exp(-ud.E1/(y(3)+273.15))*(y(1)-y(2))...
            - y(2)*ud.u1;
        
    dx(3) = -ud.delta0*(ud.k10*exp(-ud.E1/(y(3)+273.15))*(y(1)*ud.DHab...
            + y(2)*ud.DHbc) + ud.k20*exp(-ud.E2/(y(3)+273.15))*y(1)^2*ud.DHad)... 
            + ud.alpha0*(y(4)-y(3))...
            + (ud.Tin-y(3))*ud.u1;
        
    dx(4) = ud.beta0*(y(3)-y(4)) + ud.gamma0*ud.u2;
end

function xv = mat2vec(xm,M,N)
    for k = 1:N
        xv(k*M-M+1:k*M,1) = xm(:,k);
    end
end

function xm = vec2mat(xv,M,N)
    for k = 1:N
        xm(k,:) = xv((k-1)*M+1:k*M);
    end
end

%% EOF