function nmpc_disc()
%% main function for implementation of NMPC penalty method with adjoints
%% and simple discrete system model

clear all
close all
clc
format compact
format long

%% Global defs
global d;

% dimensions
d.Nu = 2;
d.Nx = 2;
d.Nz = 2;

% time
d.N = 30;
d.tf = 1;
d.Ts = d.tf/d.N;
d.t = 0:d.Ts:d.tf;

% objective function parameters
d.P = eye(d.Nx);
d.Q = [1 0 ; 0 1];
d.R = [1 0 ; 0 0.1];

% initial state
d.x0 = [1; 1];

% reference trajectory
d.xr(1,:) = [0*ones(1,floor((d.N+1)/2)) 1*ones(1,d.N+1-floor((d.N+1)/2))];
d.xr(2,:) = [0*ones(1,floor((d.N+1)/2)) 2*ones(1,d.N+1-floor((d.N+1)/2))];
d.ur = zeros(d.Nu,d.N);
for k = 1:d.N
   d.zr(:,k) =  g(d.xr(:,k),d.ur(:,k));
end

% specify constraints
d.zctol = 0;
d.xcu = [];
d.xcl = [];
d.zcu(1,:) = 0.5*ones(1,d.N);
d.zcu(2,:) = [0.7*ones(1,floor(2*d.N/3)) 0.1*ones(1,d.N-floor(2*d.N/3))];
d.zcl(1,:) = -2*ones(1,d.N);
d.zcl(2,:) = -2*ones(1,d.N);
d.ucu = [];
d.ucl = [];

% set function handles
d.J          = @obj;
d.J_uk       = @obj_uk;
d.J_xk       = @obj_xk;
d.penalty    = @l1_penalty;
d.penalty_zk = @l1_penalty_zk;
d.max_func   = @sigmoid;
d.max_func_d = @sigmoid_d;

% penalty function parameters
d.a = 50; 
my_vec = 0:1:3;

x(:,1) = d.x0;
u = zeros(d.Nu,d.N);

%% simulation
for k=1:d.N
    x(:,k+1) = f(x(:,k),u(:,k));
    z(:,k)   = g(x(:,k),u(:,k));
end

%% fminunc
uv_opt = mat2vec(d.Nu,d.N,u);
fminunc_opts = optimset('GradObj','on','TolFun',1e-1,'DerivativeCheck','off',...
                        'PlotFcns',@optimplotfval,'Diagnostics','on',...
                        'LargeScale','off');
for k = 1:length(my_vec)
    d.my = my_vec(k);
    my = d.my
    [uv_opt,J_opt] = fminunc(@model,uv_opt,fminunc_opts);
    [J_opt,Jg_opt,x_opt,z_opt] = model(vec2mat(d.Nu,d.N,uv_opt));
    fig_plot(z_opt);
end
end


%% ######################################################################
%% forward and reverse simulation, provides objfun value and gradient 
%% ######################################################################

function [J,Jg,x,z] = model(uv)
global d;
u       = vec2mat(d.Nu,d.N,uv);
x       = zeros(d.Nx,d.N+1);
z       = zeros(d.Nz,d.N);
Jg      = zeros(d.N*d.Nu,1);

% forward sim
x(:,1) = d.x0;
for k=1:d.N
    x(:,k+1) = f(x(:,k),u(:,k));
    z(:,k) = g(x(:,k),u(:,k));
end
J = d.J(x,u,z);  

% reverse sim
l = d.P*(x(:,d.N+1)-d.xr(:,d.N+1));
for k = d.N:-1:1
    Jg((k-1)*d.Nu+1:k*d.Nu,1) = d.J_uk(x(:,k),u(:,k),z(:,k),k) + Bk(x(:,k),u(:,k))'*l;
    l = d.J_xk(x(:,k),u(:,k),z(:,k),k) + Ak(x(:,k),u(:,k))'*l;
end
end

%% ######################################################################
%%  OBJECTIVE FUNCTION
%% ######################################################################

% objective function
function J = obj(x,u,z)
global d;
J = 0;
for k = 1:d.N
    J = J + 0.5*((z(:,k)-d.zr(:,k))'*d.Q*(z(:,k)-d.zr(:,k)) + (u(:,k)-d.ur(:,k))'*d.R*(u(:,k)-d.ur(:,k))) +...
        d.penalty(x(:,k),u(:,k),z(:,k),k);
end
J = J + 0.5*((x(:,d.N+1)-d.xr(:,k))'*d.P*(x(:,d.N+1)-d.xr(:,k)));
end

function J_uk = obj_uk(xk,uk,zk,k)
global d;
J_uk = Dk(xk,uk)'*(d.Q*(zk-d.zr(:,k)) + d.penalty_zk(xk,uk,zk,k)) + d.R*(uk-d.ur(:,k));
end

function J_xk = obj_xk(xk,uk,zk,k)
global d;
J_xk = Ck(xk,uk)'*(d.Q*(zk-d.zr(:,k)) + d.penalty_zk(xk,uk,zk,k));
end

%% ######################################################################
%%  Penalty functions
%% ######################################################################

% zero penalty
function P = zero_penalty(xk,uk,zk,k)
P = 0;
end
function P_zk = zero_penalty_zk(xk,uk,zk,k)
global d;
P_zk = zeros(d.Nz,1);
end

% l1 constraint penalty function
function P = l1_penalty(xk,uk,zk,k)
global d;
P = 0;
for l = 1:d.Nz
    P = P + d.my*(d.max_func(zk(l) - d.zcu(l,k)) + d.max_func(-zk(l) + d.zcl(l,k)));
    if isnan(P)
        -zk(l)
        d.max_func(-zk(l) + d.zcl(l,k))
        pause
    end
end
end

function P_zk = l1_penalty_zk(xk,uk,zk,k)
global d;
P_zk = zeros(d.Nz,1);
for l = 1:d.Nz
    P_zk(l) = d.my*(d.max_func_d(zk(l) - d.zcu(l,k)) + d.max_func_d(-zk(l) + d.zcl(l,k)));
end
end

% max functions
function m = sigmoid(x)
global d;
if -d.a*x > 10
    m = 0;
elseif -d.a*x < -10
    m = x;
else
    m = x + 1/d.a*log(1 + exp(-d.a*x));
end
end

function md = sigmoid_d(x)
global d;
if -d.a*x > 10
    md = 0;
elseif -d.a*x < -10
    md = 1;
else
    md = 1 - exp(-d.a*x)/(1 + exp(-d.a*x));
end
end

function m = exact_max(x)
    m = max(x,0);
end

function md = exact_max_d(x)
    if x < 0 
        md = 0;
    else
        md = 1;
    end
end

%% ######################################################################
%% MODEL EQUATIONS
%% ######################################################################

% system model and jacobians
function xk1 = f(xk,uk)
global d;
xk1(1) = 0.9*xk(1) + 0.9*sin(xk(2)) + uk(1);
xk1(2) = 0.9*cos(xk(2)) + uk(2);
end

function zk = g(xk,uk)
global d;
zk = xk;
end

function Ak = Ak(xk,uk)
global d;
Ak = [0.9 0.9*cos(xk(2)) ; 0 -0.9*sin(xk(2))];
end

function Bk = Bk(xk,uk)
global d;
Bk = eye(2);
end

function Ck = Ck(xk,uk)
global d;
Ck = eye(d.Nx);
end

function Dk = Dk(xk,uk)
global d;
Dk = zeros(d.Nz,d.Nu);
end

%% ######################################################################
%% Vector stacking functions
%% ######################################################################
function xm = vec2mat(M,N,xv)
    xm = zeros(M,N);
    for k = 1:N
        xm(:,k) = xv((k-1)*M+1:k*M);
    end
end

function xv = mat2vec(M,N,xm)
    xv = zeros(M*N,1);
    for k = 1:N
        xv((k-1)*M+1:k*M) = xm(:,k);
    end
end

%% EOF