function nmpc_penalty()
%% main function for implementation of NMPC penalty method with adjoints
%% and Van de Vusse reactor model

clear all
close all
clc
format compact
format short

% SUNDIALS startup
w = pwd;
sundials_path = '/Applications/MATLAB_R2008a/toolbox/sundialsTB';
cd(sundials_path);
startup_STB();
cd(w);

global d

% dimensions
d.Nu = 2;     % dimension of u
d.Nx = 4;     % dimension of x
d.Nz = 4;     % dimension of z

% time
d.N = 100;
d.tf = 1;
d.Ts = d.tf/d.N;
d.t = 0:d.Ts:d.tf;

% reactor parameters
d.alpha0  = 30.8285;
d.gamma0  = 100;
d.k10     = 1.287e12;
d.k20     = 9.043e6;
d.DHab    = 4.2;
d.DHad    = -41.85;
d.cin     = 5300;
d.beta0   = 86.688;
d.delta0  = 3.556e-4;
d.E1      = 9758.3;
d.E2      = 8560;
d.DHbc    = -11;
d.Tin     = 104.9;

% objective function parameters
Q_mask = diag([0 1 1 0],0);
R_mask = diag([1 1],0);
d.Q    = diag([1e-2 1e-1 1e1 1e1 ],0)*Q_mask;
d.R    = diag([1e1 1e1],0)*R_mask;
d.P    = d.Q*0;

% output matrices
d.C = eye(d.Nx);
d.D = zeros(d.Nz,d.Nu);

% reference trajectory
u1s1 = 8.256;  u1s2 = 18.037;
u2s1 = -6.239; u2s2 = -4.556;
x1s1 = 3517.5; x1s2 = 2985;
x2s1 = 740;    x2s2 = 960;
x3s1 = 87;     x3s2 = 106;
x4s1 = 79.8;   x4s2 = 100.7;

d.urk(1,:) = [u1s1*ones(1,floor((d.N+1)/3)) u1s1:(u1s2-u1s1)/floor((d.N+1)/3):u1s2 u1s2*ones(1,d.N-2*floor((d.N+1)/3))];
d.urk(2,:) = [u2s1*ones(1,floor((d.N+1)/3)) u2s1:(u2s2-u2s1)/floor((d.N+1)/3):u2s2 u2s2*ones(1,d.N-2*floor((d.N+1)/3))];
d.xrk(1,:) = [x1s1*ones(1,floor((d.N+1)/3)) x1s1:(x1s2-x1s1)/floor((d.N+1)/3):x1s2 x1s2*ones(1,d.N-2*floor((d.N+1)/3))];
d.xrk(2,:) = [x2s1*ones(1,floor((d.N+1)/3)) x2s1:(x2s2-x2s1)/floor((d.N+1)/3):x2s2 x2s2*ones(1,d.N-2*floor((d.N+1)/3))];
d.xrk(3,:) = [x3s1*ones(1,floor((d.N+1)/3)) x3s1:(x3s2-x3s1)/floor((d.N+1)/3):x3s2 x3s2*ones(1,d.N-2*floor((d.N+1)/3))];
d.xrk(4,:) = [x4s1*ones(1,floor((d.N+1)/3)) x4s1:(x4s2-x4s1)/floor((d.N+1)/3):x4s2 x4s2*ones(1,d.N-2*floor((d.N+1)/3))];
for k = 1:d.N
   d.zrk(:,k) = d.C*d.xrk(:,k) + d.D*d.urk(:,k);
end

% specify output constraints
z1c1u = 4000; z1c2u = 4000;
z2c1u = 800; z2c2u = 800;
z3c1u = 100;  z3c2u = 100;
z4c1u = 110;  z4c2u = 115;
z1c1l = 2000;    z1c2l = 1500;
z2c1l = 700;    z2c2l = 700;
z3c1l = 85;    z3c2l = 80;
z4c1l = 80;    z4c2l = 75;
d.zcku(1,:) = [z1c1u*ones(1,floor((d.N+1)*3/4)) z1c2u*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zcku(2,:) = [z2c1u*ones(1,floor((d.N+1)*3/4)) z2c2u*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zcku(3,:) = [z3c1u*ones(1,floor((d.N+1)*3/4)) z3c2u*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zcku(4,:) = [z4c1u*ones(1,floor((d.N+1)*3/4)) z4c2u*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zckl(1,:) = [z1c1l*ones(1,floor((d.N+1)*3/4)) z1c2l*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zckl(2,:) = [z2c1l*ones(1,floor((d.N+1)*3/4)) z2c2l*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zckl(3,:) = [z3c1l*ones(1,floor((d.N+1)*3/4)) z3c2l*ones(1,d.N+1-floor((d.N+1)*3/4))];
d.zckl(4,:) = [z4c1l*ones(1,floor((d.N+1)*3/4)) z4c2l*ones(1,d.N+1-floor((d.N+1)*3/4))];

d.zcu = [ 5000; 850; 100; 200];
d.zcl = [ -5000; -800; -200; -200];

% function handles
d.obj_grad    = @obj_grad_adj;
d.penalty     = @l1_penalty;
d.penalty_z   = @l1_penalty_z;
d.max_func    = @sigmoid;
d.max_func_z  = @sigmoid_z;
d.plot_fig    = @plot_fig;

d.a  = 10; % sigmoid max func parameter

% constant input
um(1,:) = 8.256*ones(1,d.N);
um(2,:) = -6.239*ones(1,d.N);
uv = mat2vec(d.Nu,d.N,um);

% integration tolerance
d.rel_tol = 1e-8;
d.abs_tol = 1e-8;

%% fminunc
d.my_vec = 0:10:20;
d.my_scale = [0 1 5 0];
d.x0 = [3517; 740; 87; 79.8];
fminunc_opts = optimset('GradObj','on','TolFun',1e-4,'MaxIter',30,'DerivativeCheck','off',...
                        'PlotFcns',@optimplotfval,'Diagnostics','on',...
                        'LargeScale','off');
                  
for k = 1:length(d.my_vec)
    d.my = d.my_vec(k);
    d.my
    [uv,J] = fminunc(@d.obj_grad,uv,fminunc_opts);
    [J,x] = obj(uv);
    um = vec2mat(d.Nu,d.N,uv);
    d.plot_fig(um,x)
end
end

%% ######################################################################

function [J,Jg,x] = obj_grad_adj(uv)

global d

um = vec2mat(d.Nu,d.N,uv);

% prealloc
x = zeros(d.Nx,d.N+1);
J = 0;
Jg = zeros(d.N*d.Nu,1);

% Forward integration
q0 = 0;
options = CVodeSetOptions('RelTol',d.rel_tol,...
                          'AbsTol',d.abs_tol,...
                          'LinearSolver','Dense',...
                          'JacobianFn',@vdv_J,...
                          'MaxStep',d.Ts/10);

options = CVodeSetOptions(options,...
                          'Quadratures','on',...
                          'QuadRhsFn',@vdv_q,...
                          'QuadInitcond', q0,...
                          'QuadErrControl','on',...
                          'QuadRelTol',d.rel_tol,...
                          'QuadAbsTol',d.abs_tol);
                      
CVodeMalloc(@vdv_f,0,d.x0,options);
CVadjMalloc(1000000, 'Hermite');
tic
x(:,1) = d.x0;
for k = 1:d.N
   d.u   = um(:,k);
   d.ur  = d.urk(:,k);
   d.zr  = d.zrk(:,k);
   d.zcu = d.zcku(:,k);
   d.zcl = d.zckl(:,k);   
   [status,tk1,xk1,qk1] = CVode(d.t(k+1),'Normal');
   x(:,k+1) = xk1;
end
J = qk1 + 0.5*(x(:,d.N+1)-d.xrk(:,d.N+1))'*d.P*(x(:,d.N+1)-d.xrk(:,d.N+1));
t_fwd = toc

% Reverse integration
qB1     = zeros(d.Nu,1);
lambda1 = d.P*(x(:,d.N+1)-d.xrk(:,d.N+1));

optionsB = CVodeSetOptions('RelTol',d.rel_tol,...
                           'AbsTol',d.abs_tol,...
                           'LinearSolver','Dense',...
                           'JacobianFn',@vdv_JB,...
                           'MaxStep',d.Ts);

optionsB = CVodeSetOptions(optionsB,...
                           'Quadratures','on',...
                           'QuadRhsFn',@vdv_qB,...
                           'QuadInitcond', qB1,...
                           'QuadErrControl','on',...
                           'QuadRelTol',d.rel_tol,...
                           'QuadAbsTol',d.abs_tol);
                       
CVodeMallocB(@vdv_fB, d.tf, lambda1, optionsB);

tic
for k = d.N:-1:1
    d.u = um(:,k);
    d.ur = d.urk(:,k);
    d.zr = d.zrk(:,k);
    d.zcu = d.zcku(:,k);
    d.zcl = d.zckl(:,k);
    [status,tk,lambdak,qBk] = CVodeB(d.t(k),'Normal');
    CVodeReInitB(@vdv_fB,d.t(k),lambdak,optionsB);
    Jg((k-1)*d.Nu+1:k*d.Nu) = qBk;
end
t_rwd = toc
CVodeFree;
end

%% ######################################################################

function [J,Jg,x] = obj_grad_fin(uv)

global d

Jg = zeros(d.N*d.Nu,1);
[J,x] = obj(uv);
eps = 1e-4;
for k = 1:d.N*d.Nu
    pert = eps*uv(k);
    uvp = uv;
    uvp(k) = uvp(k)+pert;
    [Jp] = obj(uvp);
    Jg(k) = (Jp-J)/pert;
end
end

%% ######################################################################

function [J,x] = obj(uv)
global d

um = vec2mat(d.Nu,d.N,uv);

% prealloc
x = zeros(d.Nx,d.N+1);
J = 0;

% Forward integration
q0 = 0;
options = CVodeSetOptions('RelTol',d.rel_tol,...
                          'AbsTol',d.abs_tol,...
                          'LinearSolver','Dense',...
                          'JacobianFn',@vdv_J,...
                          'MaxStep',d.Ts);
                      
options = CVodeSetOptions(options,...
                          'Quadratures','on',...
                          'QuadRhsFn',@vdv_q,...
                          'QuadInitcond', q0,...
                          'QuadErrControl','on',...
                          'QuadRelTol',d.rel_tol,...
                          'QuadAbsTol',d.abs_tol);
                      
CVodeMalloc(@vdv_f,0,d.x0,options);

x(:,1) = d.x0;
for k = 1:d.N
   d.u = um(:,k);
   d.ur = d.urk(:,k);
   d.zr = d.zrk(:,k);   
   [status,tk1,xk1,q] = CVode(d.t(k+1),'Normal');
   x(:,k+1) = xk1;
   J = J + q;
   CVodeReInit(@vdv_f,tk1,xk1,options);
end
J = J + 0.5*(x(:,d.N+1)-d.xrk(:,d.N+1))'*d.P*(x(:,d.N+1)-d.xrk(:,d.N+1));
CVodeFree;
end

%% ######################################################################

function xm = vec2mat(M,N,xv)
    xm = zeros(M,N);
    for k = 1:N
        xm(:,k) = xv((k-1)*M+1:k*M);
    end
end

%% ######################################################################

function xv = mat2vec(M,N,xm)
    xv = zeros(M*N,1);
    for k = 1:N
        xv((k-1)*M+1:k*M) = xm(:,k);
    end
end

%% EOF