function obj_grad()
%% Adjoint and forward objective function gradient calculation

clear all
close all
clc
format compact
format short

%% SET OPTIONS (true = 1, false = 0)

run = struct;
run.plot_fig         = 1; % plot solution
run.plot_fig_title   = 1; % title names in figures
run.plot_runtime_fig = 1; % plot runtime figures
run.rt_mode          = 2; % choose runtime measurement mode

pts = 100;     % number of points to measure runtime
avg_pts = 100; % number of iterations to average over

% different modes for measuring runtime
switch run.rt_mode
    case 1
        rt_N = 1:5:pts;
    case 2
        rt_N  = 10;        % time horizon
        avg_pts = 1;
    otherwise
        fprintf('ERROR: rt_mode not valid');
end

%% init
Ts = 1;         % sampling time
Nx = 2;         % number of states
Nu = 2;         % number of inputs
Nz = 1;         % number of outputs

for avg_iter = 1:avg_pts;
    for rt_iter = 1:length(rt_N);
        
        fprintf(['Iteration ' num2str(rt_iter) '/' num2str(pts) ...
            ', Average Iteration ' num2str(avg_iter) '/' num2str(avg_pts) '\n']);

        % time
        N = rt_N(rt_iter);

        % prealloc for speed
        x = zeros(Nx,N+1);  % states
        u = zeros(Nu,N);    % controlled variables
        z = zeros(Nz,N);    % controlled outputs

        % initial value and time
        t0 = 1;                         % start time
        x(:,1)=[1;1];                   % initial state
        t = (0:N)*Ts + t0;              % time vector

        %% simulation with feedback linearization controller
        for k=1:N
            u(1,k)=-x(2,k)^2;
            u(2,k)= x(1,k)*t(k);
            [x(:,k+1),z(:,k)] = f(x(:,k),u(:,k),t(k));
        end

        if run.plot_fig == true
            
        %% FIGURES
            
            figure()
            for k = 1:Nx
                plot(t,x(k,:))
                hold all
            end
            grid
            xlabel('k')
            ylabel('x')
            if run.plot_fig_title == true
                title('states [x] (feedback linearization controller)')
            end

            figure()
            for k = 1:Nz
                plot(t(1:N),z(k,:))
                hold all
            end
            xlabel('k')
            ylabel('z')
            if run.plot_fig_title == true
                title('output [z] (feedback linearization controller)')
            end
            grid

            figure()
            for k = 1:Nu
                plot(t(1:N),u(k,:))
                hold all
            end
            xlabel('k')
            ylabel('u')
            legend('u_1','u_2')
            if run.plot_fig_title == true
                title('control [u] (feedback linearization controller)')
            end
            grid
        end

        % objective function parameters - sum(xk'*Q*xk + uk*'R*uk)/2 + xN'*P*xN/2
        P = eye(Nx);
        Q = eye(Nz);
        R = eye(Nu);

        %% Finite differences
        tic;
        GradJ_fin = zeros(N,Nu);    % prealloc for speed
        J0 = zeros(N+2,1);
        x_diff = zeros(Nx,N);
        z_diff = zeros(Nz,N);

        for k = 1:N
            J0(k+1) = J0(k) + z(:,k)'*Q*z(:,k)/2 + u(:,k)'*R*u(:,k)/2;
        end
        J0(N+2) = J0(N+1) + x(:,N+1)'*P*x(:,N+1)/2;

        eps = 1e-10;
        for k = 1:N
            for l = 1:Nu
                Jkl = J0(k);
                u_diff = u;
                u_diff(l,k) = u_diff(l,k) + eps;
                x_diff(:,k) = x(:,k);
                for n = k:N
                    [x_diff(:,n+1),z_diff(:,n)] = f(x_diff(:,n),u_diff(:,n),t(n));
                    Jkl = Jkl + z_diff(:,n)'*Q*z_diff(:,n)/2 + u_diff(:,n)'*R*u_diff(:,n)/2;
                end
                Jkl = Jkl + x_diff(:,N+1)'*P*x_diff(:,N+1)/2;
                GradJ_fin(k,l) = (Jkl-J0(N+2))/eps;
            end
        end
        cpu_time_fin(rt_iter,avg_iter) = toc;

        %% Forward sensitivity compuatation
        tic;
        GradJ_fwd = zeros(N,Nu);    % prealloc for speed
        for k = 1:N-1
            Atk = eye(Nx);          % placeholder for A(i-1)*...*A(k+1)
            ZQCA = zeros(1,Nx);
            for i = k+2:N
                Atk = fx(x(:,i-1),u(:,i-1),t(i-1))*Atk;
                ZQCA = ZQCA + z(:,i)'*Q*gx(x(:,i),u(:,i),t(i))*Atk;
            end

            GradJ_fwd(k,:) = u(:,k)'*R + z(:,k)'*Q*gu(x(:,k),u(:,k),t(k)) + ...
                (z(:,k+1)'*Q*gx(x(:,k+1),u(:,k+1),t(k+1)) + ZQCA + ...
                x(:,N+1)'*P*fx(x(:,N),u(:,N),t(N))*Atk)*fu(x(:,k),u(:,k),t(k));

        end
        GradJ_fwd(N,:) = u(:,N)'*R + z(:,N)'*Q*gu(x(:,N),u(:,N),t(N)) + ...
            x(:,N+1)'*P*fu(x(:,N),u(:,N),t(N));
        cpu_time_fwd(rt_iter,avg_iter) = toc;

        %% Adjoint sensitivity computation (reverse)
        tic
        GradJ_adj = zeros(N,Nu);
        lambda = zeros(Nx,N);
        lambda(:,N+1) = P*x(:,N+1);

        for k = N:-1:1
            lambda(:,k) = gx(x(:,k),u(:,k),t(k))'*Q*z(:,k) + ...
                fx(x(:,k),u(:,k),t(k))'*lambda(:,k+1);
        end

        for k = 1:N
            GradJ_adj(k,:) = (R*u(:,k)+ gu(x(:,k),u(:,k),t(k))'*Q*z(:,k) + ...
                fu(x(:,k),u(:,k),t(k))'*lambda(:,k+1))';
        end
        cpu_time_adj(rt_iter,avg_iter) = toc;
    end
end
if run.plot_runtime_fig == true

    % calculate average of all iterations
    cpu_time_fin_avg = zeros(length(rt_N),1);
    cpu_time_fwd_avg = zeros(length(rt_N),1);
    cpu_time_adj_avg = zeros(length(rt_N),1);
    for k = 1:avg_pts
        cpu_time_fin_avg = cpu_time_fin_avg + cpu_time_fin(:,k);
        cpu_time_fwd_avg = cpu_time_fwd_avg + cpu_time_fwd(:,k);
        cpu_time_adj_avg = cpu_time_adj_avg + cpu_time_adj(:,k);        
    end
    cpu_time_fin_avg = cpu_time_fin_avg/avg_pts;
    cpu_time_fwd_avg = cpu_time_fwd_avg/avg_pts;
    cpu_time_adj_avg = cpu_time_adj_avg/avg_pts;

    %% RUNTIME FIGURES

    switch run.rt_mode
        case 1
            % N variable
            figure()
            plot(rt_N,cpu_time_fin_avg,'-*');
            hold all
            plot(rt_N,cpu_time_fwd_avg,'-*');
            hold all
            plot(rt_N,cpu_time_adj_avg,'-*');
            if run.plot_fig_title == true
                title(['Obj_grad Runtime']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            legend('Finite differences','Forward','Adjoint');
            grid;

            figure()
            plot(rt_N,cpu_time_adj_avg,'-*');
            if run.plot_fig_title == true
                title(['Obj_grad Runtime - adjoint']);
            end
            xlabel('N');
            ylabel('CPU time [s]');
            grid;            
        case 2
            % does not apply to this mode
        otherwise
            fprintf('ERROR: rt_mode not valid');
    end
end
end

%% SECIFY SYTEM MODEL - f(xk,uk,tk) and jacobians Ak, Bk, Ck, Dk

function [xk1,zk] = f(xk,uk,tk)

xk1(1) = xk(1)/tk + xk(2)^2 + uk(1);
xk1(2) = xk(2)*sin(tk) - xk(1)*tk + uk(2);

zk = xk(1) +  xk(2) + 0.1*(uk(1) + 2*uk(2));
end

function [Ak] = fx(xk,uk,tk) % return jacobian df(x,u,t(k))/dx
Ak = [1/tk 2*xk(2); -tk sin(tk)];
end

function [Bk] = fu(xk,uk,tk) % return jacobian df(x,u,t(k))/du
Bk = [1 0; 0 1];
end

function [Ck] = gx(xk,uk,tk) % return jacobian dg(x,u,t(k))/dx
Ck = [1 1];
end

function [Dk] = gu(xk,uk,tk) % return jacobian df(x,u,t(k))/du
Dk = [0.1 0.2];
end

%% EOF