function vdv_ss()
%% Calculates steady state solution of Van de Vusse reactor

clear all
close all
clc
format compact
format short

%% SET RUN OPTIONS (true = 1, false = 0)
run.plot_fig_title   = 1; % title names in figures

%% reactor parameters
global alpha gamma0 k10 k20 DHab DHad...
       cin beta0 delta E1 E2 DHbc Tin;

alpha   = 30.8285;
gamma0  = 0.1;
k10     = 1.287e12;
k20     = 9.043e6;
DHab    = 4.2;
DHad    = -41.85;
cin     = 5100;
beta0   = 86.688;
delta   = 3.556e-4;
E1      = 9758.3;
E2      = 8560;
DHbc    = -11;
Tin     = 104.9;

global u1 u2;

x0 = [1000 1000 0 1000];

u1min = 5; 
u1max = 35;
u1step = 1;
u2min = -8.5; % MJ
u2max = 0;    % MJ
u2step = 0.25;

u1grid = u1min:u1step:u1max;
u2grid = u2min:u2step:u2max;

cbgrid = zeros(length(u1grid),length(u2grid));

for k = 1:length(u1grid)    
    for n = 1:length(u2grid)
        u1 = u1grid(k);
        u2 = u2grid(n)*1e3;
        [X] = fsolve(@van_de_vusse,x0);

        cagrid(k,n,:) = X(1);
        cbgrid(k,n,:) = X(2);
        Tgrid(k,n,:) = X(3);
        Tcgrid(k,n,:) = X(4);        
    end
end


figure()
mesh(u2grid,u1grid,cagrid)
if run.plot_fig_title == true
    title('c_A')
end
zlabel('c_A [mol/m^3]');
ylabel('u_1 [1/h]');
xlabel('u_2 [MJ]');
colormap jet

figure()
mesh(u2grid,u1grid,cbgrid)
if run.plot_fig_title == true
    title('c_B')
end
zlabel('c_B [mol/m^3]');
ylabel('u_1 [1/h]');
xlabel('u_2 [MJ]');
colormap jet

figure()
mesh(u2grid,u1grid,Tgrid)
if run.plot_fig_title == true
    title('T')
end    
zlabel('T [C]');
ylabel('u_1 [1/h]');
xlabel('u_2 [MJ]');
colormap jet

figure()
mesh(u2grid,u1grid,Tcgrid)
if run.plot_fig_title == true
    title('T_c')
end
zlabel('T_c [C]');
ylabel('u_1 [1/h]');
xlabel('u_2 [MJ]');
colormap jet

end

function dx = van_de_vusse(x)
global alpha gamma0 cin beta0 Tin u1 u2;

    ca = x(1);    
    cb = x(2);
    T  = x(3);
    Tc = x(4);    
    
    dx = zeros(4,1);
    dx(1) = -k1(T)*ca - k2(T)*ca^2 + (cin-ca)*u1;
    dx(2) = k1(T)*(ca-cb) - cb*u1;
    dx(3) = h(ca,cb,T) + alpha*(Tc-T) + (Tin-T)*u1;
    dx(4) = beta0*(T-Tc) + gamma0*u2;
end


function k1 = k1(T)
    global k10 E1;
    k1 = k10*exp(-E1/(T+273.15));
end

function k2 = k2(T)
    global k20 E2;
    k2 = k20*exp(-E2/(T+273.15));
end

function h = h(ca,cb,T)
    global DHab DHad delta DHbc;
    h = -delta*(k1(T)*(ca*DHab + cb*DHbc) + k2(T)*ca^2*DHad);
end

%% EOF
